/*
// A simple C++-style interface for polynomials
//
//
//
*/

#include "poly.h"
#include <cmath>

// use this command to avoiud std::cout
using namespace std ;

/*
//  static = visibility only in this file
//  inline = expand inline the function when called
//
//  expression ? execute if true : execute if false ;
*/

static
inline
indexType
min( indexType a, indexType b )
{ return a < b ? a : b ; }
// equivalent
//
// if ( a < b ) 
//   return a ;
// else
//   return b ;
//

static
inline
indexType
max( indexType a, indexType b )
{ return a > b ? a : b ; }

/*
//       _ _       
//    __| (_)_   __
//   / _` | \ \ / /
//  | (_| | |\ V / 
//   \__,_|_| \_/  
*/

// divide two polynomial a/b return the quotient and the rest
// a = p * b + r
void
div( Poly const & a, Poly const & b, Poly & p, Poly & r ) {
  // compute true order of a and b
  indexType aOrder = a.getOrder() ;
  indexType bOrder = b.getOrder() ;
  
  if ( bOrder < 0 ) {
    cerr << "div division by zero\n" ;
    exit(1) ; // error found, exiting to the program
  }
  
  r = a ;
  if ( bOrder > aOrder ) {
    // if order b > order a then
    // p = 0 and r = a
    p = 0 ;
    return ;
  }
  
  p.resize(aOrder-bOrder) ;
  for ( indexType i = aOrder ; i >= bOrder ; --i ) {
    // compute coeff to erase a.coeffs[i]
    valueType cf = r[i] / b[bOrder] ;
    for ( indexType j = 0 ; j < bOrder ; ++j )
      r[i+j-bOrder] -= cf * b[j] ;
    r[i]        = 0  ;
    p[i-bOrder] = cf ;
  }
}

/*
//                     _ 
//   _ __ ___   ___ __| |
//  | '_ ` _ \ / __/ _` |
//  | | | | | | (_| (_| |
//  |_| |_| |_|\___\__,_|
*/
// compute Maximum Common Divisor
void
mcd( Poly const & p,
     Poly const & q,
     indexType  & nPoly,
     Poly         polys[] ) {

  // compute true order of p and r
  indexType pOrder = p.getOrder() ;
  indexType qOrder = q.getOrder() ;

  if ( pOrder > qOrder ) {
    polys[0] = p ;
    polys[1] = q ;
  } else {
    polys[0] = q ;
    polys[1] = p ;
  }

  Poly buffer ;
  for ( nPoly=2 ; polys[nPoly-1] . getOrder() > 0 ; ++nPoly ) {
    div( polys[nPoly-2],
         polys[nPoly-1],
         buffer,
         polys[nPoly] ) ;
    polys[nPoly] . drop(1E-10) ;
    polys[nPoly] = -polys[nPoly] ;
  }
}

/*       _ _  __  __ 
//    __| (_)/ _|/ _|
//   / _` | | |_| |_ 
//  | (_| | |  _|  _|
//   \__,_|_|_| |_|  
*/
// compute polynomial derivative
Poly
diff( Poly const & p ) {
  indexType pOrder = p . getOrder() ;

  if ( pOrder <= 0 ) return Poly(0) ;
  
  Poly res( pOrder - 1 ) ;
  for ( indexType i = 0 ; i < pOrder ; ++i )
    res[i] = p[i+1]*(i+1) ;

  return res ;
}

/*
//   ____  _                        
//  / ___|| |_ _   _ _ __ _ __ ___  
//  \___ \| __| | | | '__| '_ ` _ \ 
//   ___) | |_| |_| | |  | | | | | |
//  |____/ \__|\__,_|_|  |_| |_| |_|
*/
// compute Sturm Sequence
void
Sturm( Poly const & p, indexType & nPoly, Poly polys[] ) {
  Poly p1 = diff(p) ; // compute polynomial derivative
  mcd( p, p1, nPoly, polys ) ; // compute MCP between p and p1

  // divide polynomials sequence by MCD and make polynomials monic
  Poly A, R ;
  for ( indexType i = 0 ; i < nPoly ; ++i ) {
    // divide two polynomial a/b return the quotient and the rest
    // a = p * b + r
    div( polys[i], polys[nPoly-1], A, R ) ;
    polys[i] = A / abs(A.leading()) ;
  }
}

/*   ____  _                         ____ _                                 
//  / ___|| |_ _   _ _ __ _ __ ___  / ___| |__   __ _ _ __   __ _  ___  ___ 
//  \___ \| __| | | | '__| '_ ` _ \| |   | '_ \ / _` | '_ \ / _` |/ _ \/ __|
//   ___) | |_| |_| | |  | | | | | | |___| | | | (_| | | | | (_| |  __/\__ \
//  |____/ \__|\__,_|_|  |_| |_| |_|\____|_| |_|\__,_|_| |_|\__, |\___||___/
//                                                          |___/           
*/
// compute sign changes
indexType
SturmChanges( indexType  nPoly,
              Poly const polys[],
              valueType  epsi,
              valueType  x ) {
  indexType nChange = 0 ;
  valueType v = polys[0](x) ;
  bool      p = v > epsi ;
  for ( indexType i = 1 ; i < nPoly ; ++i ) {
    valueType newv = polys[i](x) ;
    if ( abs(newv) < epsi ) continue ; // skip

    bool newp = newv > epsi ;
    if ( p != newp ) ++nChange ;
    p = newp ;
    v = newv ;
  }
  return nChange ;
}

/*
//   ____  _                         ____ _                                 
//  / ___|| |_ _   _ _ __ _ __ ___  / ___| |__   __ _ _ __   __ _  ___  ___ 
//  \___ \| __| | | | '__| '_ ` _ \| |   | '_ \ / _` | '_ \ / _` |/ _ \/ __|
//   ___) | |_| |_| | |  | | | | | | |___| | | | (_| | | | | (_| |  __/\__ \
//  |____/ \__|\__,_|_|  |_| |_| |_|\____|_| |_|\__,_|_| |_|\__, |\___||___/
//                                                          |___/           
*/

void
SturmChanges( ostream & s,
              indexType  nPoly,
              Poly const polys[],
              valueType  epsi,
              valueType  x ) {
  valueType v = polys[0](x) ;
  if ( v > epsi )       s << "[ + " ;
  else if ( v < -epsi ) s << "[ - " ;
  else                  s << "[ 0 " ;
  for ( indexType i = 1 ; i < nPoly ; ++i ) {
    valueType v = polys[i](x) ;
    if ( v > epsi )       s << ", + " ;
    else if ( v < -epsi ) s << ", - " ;
    else                  s << ", 0 " ;
  }
  s << "]\n" ;
}

#if 0
// compute polynomial integral
void
PolyInt( Poly const & p, Poly & q )  {
  PolyDestroy(q) ;
  PolyCreate(q,p.order+1) ;
  q.coeffs[0] = 0 ;
  for ( indexType i = 0 ; i <= p.order ; ++i )
    q.coeffs[i+1] = p.coeffs[i]/(i+1) ;
}
#endif

